﻿// GameDraw.cs

using System;
using System.Collections.Generic;
//using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
//using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
using System.Diagnostics;



/******************************************************************************/
/*
    GameDraw Class
    All the basic draw logic was kept in this class so it can easily be
    optimised or updated
*/
/******************************************************************************/
public class GameDraw
{
    public GraphicsDeviceManager    graphics;
    public Vector2                  screenSize;
    LineDraw2D                      lineDraw2D;
    LineDraw3D                      lineDraw3D;
    TriangleDraw2D                  triangleDraw2D;
    TriangleDraw3D                  triangleDraw3D;
    TexTriangleDraw3D               texTriangle3D;

    public GameDraw(GraphicsDeviceManager graphics)
    {
        this.graphics      = graphics;
        screenSize.X = (float)(graphics.GraphicsDevice.Viewport.Width);
        screenSize.Y = (float)(graphics.GraphicsDevice.Viewport.Height);

        lineDraw2D         = new LineDraw2D(graphics.GraphicsDevice, 1.0f);
        triangleDraw2D     = new TriangleDraw2D(graphics.GraphicsDevice, 1.0f);
        triangleDraw3D     = new TriangleDraw3D(graphics.GraphicsDevice, 1.0f);
        texTriangle3D      = new TexTriangleDraw3D(graphics.GraphicsDevice, 1.0f);
        lineDraw3D         = new LineDraw3D(graphics.GraphicsDevice, 1.0f);
    }

    public void DrawColourRect2D(Vector2 pos, Vector2 size, Color c, float z)
    {
        triangleDraw2D.Batch(pos, new Vector2(pos.X + size.X, pos.Y), pos + size, c, z);
        triangleDraw2D.Batch(pos, pos + size, new Vector2(pos.X, pos.Y + size.Y), c, z);
    }

    public void DrawTriangle2D(Vector2 p0, Vector2 p1, Vector2 p2, Color c, float z)
    {
        triangleDraw2D.Batch(p0, p1, p2, c, z);
    }
    
    public void DrawTriangle3D(Vector3 p0, Vector3 p1, Vector3 p2, Color c)
    {
        triangleDraw3D.Batch(p0,p1,p2,c);
    }
    
    public void DrawTexTriangle3D(VertexPositionTexture[] p, Texture2D tex)
    {
        Debug.Assert(p.Length==3);
        texTriangle3D.Batch(tex, p[0], p[1], p[2]);
    }

    public void DrawLine2D(Vector2 from, Vector2 to)
    {
        lineDraw2D.Batch(from, to, Color.Black, 0);
    }
    
    public void DrawLine3D(Vector3 from, Vector3 to, Color c)
    {
        lineDraw3D.Batch(from, to, c);
    }
    
    public void DrawCross3D(Vector3 pos, Color c)
    {
        DrawLine3D( pos+new Vector3(0,-1,0), pos+new Vector3(0,1,0), c );
        DrawLine3D( pos+new Vector3(-1,0,0), pos+new Vector3(1,0,0), c );
        DrawLine3D( pos+new Vector3(0,0,-1), pos+new Vector3(0,0,1), c );
    }

    public void DrawWireBox2D(float x, float y, float w, float h, Color c, float z)
    {
        lineDraw2D.Batch(new Vector2(x, y), new Vector2(x + w, y), c, z);
        lineDraw2D.Batch(new Vector2(x + w, y), new Vector2(x + w, y + h), c, z);
        lineDraw2D.Batch(new Vector2(x + w, y + h), new Vector2(x, y + h), c, z);
        lineDraw2D.Batch(new Vector2(x, y + h), new Vector2(x, y), c, z);
    }

    public void Begin()
    {
        triangleDraw3D.Begin();
        lineDraw2D.Begin();
        triangleDraw2D.Begin();
        texTriangle3D.Begin();
        lineDraw3D.Begin();
    }

    public void End()
    {
        triangleDraw3D.End();
        lineDraw2D.End();
        triangleDraw2D.End();
        texTriangle3D.End();
        lineDraw3D.End();
    }
};

/******************************************************************************/
/*
    LineDraw2D Class
    Used by our GameDraw above to keep all our line drawing down to a single
    big render
*/
/******************************************************************************/
class LineDraw2D
{
    private GraphicsDevice              graphicsDevice;
    private List<VertexPositionColor>   points = new List<VertexPositionColor>();
    private VertexDeclaration           vertexDeclaration;
    private BasicEffect                 basicEffect;

    public LineDraw2D(GraphicsDevice graphicsDevice, float alpha)
    {
        this.graphicsDevice = graphicsDevice;

        basicEffect = new BasicEffect( graphicsDevice, null );
        basicEffect.VertexColorEnabled = true;
        basicEffect.Alpha = alpha;
        basicEffect.Projection = Matrix.CreateOrthographicOffCenter( 0.0F,
            graphicsDevice.Viewport.Width, graphicsDevice.Viewport.Height, 0.0F,
            0.0F, -1.0F );
        basicEffect.View  = Matrix.Identity;
        basicEffect.World = Matrix.Identity;

        vertexDeclaration = 
        new VertexDeclaration( graphicsDevice, VertexPositionColor.VertexElements );
    }

    public void Begin()
    {
        points.Clear();
    }

    public void Batch( Vector2 startPoint, Vector2 endPoint,
        Color color, float layerDepth )
    {
        Batch( startPoint, color, layerDepth );
        Batch( endPoint, color, layerDepth );
    }

    public void Batch( Vector2 startPoint, Color startColor, 
        Vector2 endPoint, Color endColor, float layerDepth )
    {
        Batch( startPoint, startColor, layerDepth );
        Batch( endPoint, endColor, layerDepth );
    }

    public void Batch( Vector2 point, Color color, float layerDepth )
    {
        VertexPositionColor batchPoint =
            new VertexPositionColor(
            new Vector3( point, layerDepth ), color );
        points.Add( batchPoint );
    }

    public void End()
    {
        if( points.Count > 0 )
        {
            graphicsDevice.VertexDeclaration = vertexDeclaration;
            basicEffect.Begin();

            foreach( EffectPass effectPass in basicEffect.CurrentTechnique.Passes )
            {
                effectPass.Begin();

                int numLines = points.Count;
                if (numLines>30000) numLines=30000;

                graphicsDevice.DrawUserPrimitives<VertexPositionColor>(
                    PrimitiveType.LineList, points.ToArray(), 0, numLines / 2);

                effectPass.End();
            }

            basicEffect.End();
        }
    }
}

/******************************************************************************/
/*
    LineDraw3D Class
    Used by our GameDraw above to keep all our line drawing down to a single
    big render
*/
/******************************************************************************/
class LineDraw3D
{
    private GraphicsDevice              graphicsDevice;
    private List<VertexPositionColor>   points = new List<VertexPositionColor>();
    private VertexDeclaration           vertexDeclaration;
    private BasicEffect                 basicEffect;

    public LineDraw3D(GraphicsDevice graphicsDevice, float alpha)
    {
        this.graphicsDevice = graphicsDevice;

        basicEffect = new BasicEffect( graphicsDevice, null );
        basicEffect.VertexColorEnabled = true;
        basicEffect.Alpha = alpha;
        basicEffect.Projection = Matrix.CreatePerspective(graphicsDevice.Viewport.Width / graphicsDevice.Viewport.Height, 
                                                          1.0f,
                                                          1.0f,
                                                          1000.0f);
        basicEffect.View  = Matrix.CreateLookAt(new Vector3(0,10,-200),
                                                new Vector3(0,0,0),
                                                new Vector3(0,1,0));
        basicEffect.World = Matrix.Identity;

        vertexDeclaration = 
        new VertexDeclaration( graphicsDevice, VertexPositionColor.VertexElements );
    }

    public void Begin()
    {
        points.Clear();
    }

    public void Batch( Vector3 startPoint, Vector3 endPoint, Color color )
    {
        Batch( startPoint, color );
        Batch( endPoint,   color );
    }

    public void Batch( Vector3 point, Color color )
    {
        VertexPositionColor batchPoint = new VertexPositionColor(point, color );
        points.Add( batchPoint );
    }

    public void End()
    {
        if( points.Count > 0 )
        {
            graphicsDevice.VertexDeclaration = vertexDeclaration;
            basicEffect.Begin();

            foreach( EffectPass effectPass in basicEffect.CurrentTechnique.Passes )
            {
                effectPass.Begin();

                int numLines = points.Count;
                if (numLines>30000) numLines=30000;

                graphicsDevice.DrawUserPrimitives<VertexPositionColor>(
                    PrimitiveType.LineList, points.ToArray(), 0, numLines / 2);

                effectPass.End();
            }

            basicEffect.End();
        }
    }
}

/******************************************************************************/
/*
    TriangleDraw2D Class
    So we can group together drawing lots of triangles in one go!
*/
/******************************************************************************/
class TriangleDraw2D
{
    private GraphicsDevice              graphicsDevice;
    private List<VertexPositionColor>   points = new List<VertexPositionColor>();
    private VertexDeclaration           vertexDeclaration;
    private BasicEffect                 basicEffect;

    public TriangleDraw2D(GraphicsDevice graphicsDevice, float alpha)
    {
        this.graphicsDevice = graphicsDevice;

        basicEffect = new BasicEffect( graphicsDevice, null );
        basicEffect.VertexColorEnabled = true;
        basicEffect.Alpha = alpha;
        basicEffect.Texture = null;
        basicEffect.TextureEnabled = false;
        basicEffect.Projection = Matrix.CreateOrthographicOffCenter( 0.0F,
            graphicsDevice.Viewport.Width, graphicsDevice.Viewport.Height, 0.0F,
            0.0F, -1.0F );
        basicEffect.View  = Matrix.Identity;
        basicEffect.World = Matrix.Identity;

        vertexDeclaration = 
        new VertexDeclaration( graphicsDevice, VertexPositionColor.VertexElements );
    }

    public void Begin()
    {
        points.Clear();
    }

    public void Batch( Vector2 p0, Vector2 p1, Vector2 p2, Color color, float layerDepth)
    {
        VertexPositionColor v0 = new VertexPositionColor(new Vector3( p0, layerDepth ), color );
        VertexPositionColor v1 = new VertexPositionColor(new Vector3( p1, layerDepth ), color );
        VertexPositionColor v2 = new VertexPositionColor(new Vector3( p2, layerDepth ), color );
        points.Add( v0 );
        points.Add( v1 );
        points.Add( v2 );
    }

    public void End()
    {
        if( points.Count > 0 )
        {
            //graphicsDevice.SamplerStates[0].MagFilter = TextureFilter.None;
            //graphicsDevice.SamplerStates[0].MipFilter = TextureFilter.None;
            //graphicsDevice.SamplerStates[0].MinFilter = TextureFilter.None;
            //graphicsDevice.RenderState.MultiSampleAntiAlias = false;
            //graphicsDevice.RenderState.PointSize = 1.0f;

            graphicsDevice.VertexDeclaration        = vertexDeclaration;
            //graphicsDevice.RenderState.FillMode     = FillMode.Solid;
            //graphicsDevice.RenderState.CullMode     = CullMode.CullCounterClockwiseFace;

            basicEffect.Begin();

            foreach( EffectPass effectPass in basicEffect.CurrentTechnique.Passes )
            {
                effectPass.Begin();

                graphicsDevice.DrawUserPrimitives<VertexPositionColor>(
                    PrimitiveType.TriangleList, points.ToArray(), 0, points.Count/3);

                effectPass.End();
            }

            basicEffect.End();
        }
    }
}



/******************************************************************************/
/*
    TriangleDraw3D Class
    So we can group together drawing lots of triangles in one go!
*/
/******************************************************************************/
class TriangleDraw3D
{
    private GraphicsDevice graphicsDevice;
    private List<VertexPositionColor> points = new List<VertexPositionColor>();
    private VertexDeclaration vertexDeclaration;
    private BasicEffect basicEffect;

    public TriangleDraw3D(GraphicsDevice graphicsDevice, float alpha)
    {
        this.graphicsDevice = graphicsDevice;

        basicEffect = new BasicEffect(graphicsDevice, null);
        basicEffect.VertexColorEnabled = true;
        basicEffect.Alpha = alpha;
        basicEffect.Texture = null;
        basicEffect.TextureEnabled = false;
        basicEffect.Projection = Matrix.CreatePerspective(graphicsDevice.Viewport.Width / graphicsDevice.Viewport.Height, 
                                                          1.0f,
                                                          1.0f,
                                                          1000.0f);
        basicEffect.View  = Matrix.CreateLookAt(new Vector3(0,10,-200),
                                                new Vector3(0,0,0),
                                                new Vector3(0,1,0));
        basicEffect.World = Matrix.Identity;


        vertexDeclaration =
        new VertexDeclaration(graphicsDevice, VertexPositionColor.VertexElements);
    }

    public void Begin()
    {
        points.Clear();
    }

    public void Batch(Vector3 p0, Vector3 p1, Vector3 p2, Color color)
    {
        VertexPositionColor v0 = new VertexPositionColor(p0, color);
        VertexPositionColor v1 = new VertexPositionColor(p1, color);
        VertexPositionColor v2 = new VertexPositionColor(p2, color);
        points.Add(v0);
        points.Add(v1);
        points.Add(v2);
    }

    public void End()
    {
        if (points.Count > 0)
        {
            graphicsDevice.VertexDeclaration = vertexDeclaration;
            basicEffect.Begin();

            foreach (EffectPass effectPass in basicEffect.CurrentTechnique.Passes)
            {
                effectPass.Begin();

                graphicsDevice.DrawUserPrimitives<VertexPositionColor>(
                    PrimitiveType.TriangleList, points.ToArray(), 0, points.Count / 3);

                effectPass.End();
            }

            basicEffect.End();
        }
    }
}


/******************************************************************************/
/*
    TexTriangleDraw3D Class
    So we can group together drawing lots of triangles in one go!
*/
/******************************************************************************/
class TexTriangleDraw3D
{
    public class TexPoints
    {
        public Texture2D                    tex    = null;
        public List<VertexPositionTexture>  points = new List<VertexPositionTexture>();
    };

    private GraphicsDevice              graphicsDevice;
    private List<TexPoints>             texPoints             = new List<TexPoints>();
    private VertexDeclaration           vertexDeclaration;
    private BasicEffect                 basicEffect;

    public TexTriangleDraw3D( GraphicsDevice graphicsDevice, float alpha )
    {
        this.graphicsDevice = graphicsDevice;

        basicEffect = new BasicEffect( graphicsDevice, null );
        basicEffect.VertexColorEnabled  = false;
        basicEffect.Alpha               = alpha;
        basicEffect.Texture             = null;
        basicEffect.TextureEnabled      = true;
        basicEffect.EnableDefaultLighting();
        basicEffect.LightingEnabled = false;
        basicEffect.Projection = Matrix.CreatePerspective(graphicsDevice.Viewport.Width / graphicsDevice.Viewport.Height, 
                                                          1.0f,
                                                          1.0f,
                                                          1000.0f);
        basicEffect.View  = Matrix.CreateLookAt(new Vector3(0,10,-200),
                                                new Vector3(0,0,0),
                                                new Vector3(0,1,0));
        basicEffect.World = Matrix.Identity;

        vertexDeclaration = 
        new VertexDeclaration( graphicsDevice, VertexPositionTexture.VertexElements );
    }

    public void Begin()
    {
        for (int i=0; i<texPoints.Count; i++)
        {
            texPoints[i].tex = null;
            texPoints[i].points.Clear();
        }
        texPoints.Clear();
    }

    public void Batch(Texture2D tex, VertexPositionTexture v0, 
                                     VertexPositionTexture v1, 
                                     VertexPositionTexture v2)
    {
        int found = -1;
        for (int i=0; i<texPoints.Count; i++)
        {
            if (tex==texPoints[i].tex)
            {
                found = i;
                break;
            }
        }
        if (found>-1)
        {
            texPoints[found].points.Add( v0 );
            texPoints[found].points.Add( v1 );
            texPoints[found].points.Add( v2 );
        }
        else
        {
            texPoints.Add( new TexPoints() );
            texPoints[ texPoints.Count-1 ].tex = tex;
            texPoints[ texPoints.Count-1 ].points.Add( v0 );
            texPoints[ texPoints.Count-1 ].points.Add( v1 );
            texPoints[ texPoints.Count-1 ].points.Add( v2 );
        }
    }

    public void End()
    {
        if( texPoints.Count <= 0 ) return;

        graphicsDevice.VertexDeclaration        = vertexDeclaration;
        //graphicsDevice.RenderState.FillMode     = FillMode.Solid;
        //graphicsDevice.RenderState.CullMode     = CullMode.CullCounterClockwiseFace;
        graphicsDevice.SamplerStates[0].AddressU   = TextureAddressMode.Border;   
        graphicsDevice.SamplerStates[0].AddressV   = TextureAddressMode.Border; 


        graphicsDevice.RenderState.AlphaBlendEnable = true;            
        graphicsDevice.RenderState.SourceBlend = Blend.SourceAlpha ;            
        graphicsDevice.RenderState.DestinationBlend = Blend.InverseSourceAlpha;            
        graphicsDevice.RenderState.DepthBufferWriteEnable = true;


        for (int i=0; i<texPoints.Count; i++)
        {
            basicEffect.Texture = texPoints[i].tex;

            basicEffect.Begin();

            foreach( EffectPass effectPass in basicEffect.CurrentTechnique.Passes )
            {
                effectPass.Begin();

                int numTris                     = texPoints[i].points.Count/3;
                int startTri                    = 0;
                int numTrisToDraw               = numTris;
                int maxTrisDraw                 = 5000;
                VertexPositionTexture[] verts   = texPoints[i].points.ToArray();
                
                while (numTrisToDraw>0)
                {
                    if (numTrisToDraw > maxTrisDraw) numTrisToDraw = maxTrisDraw;
                    
                    graphicsDevice.DrawUserPrimitives<VertexPositionTexture>(PrimitiveType.TriangleList, 
                                                                             verts, 
                                                                             startTri*3, 
                                                                             numTrisToDraw);

                    startTri += maxTrisDraw;
                    numTrisToDraw = (numTris - startTri);
                    
                }

                effectPass.End();
            }

            basicEffect.End();
        }
    }
}






